@@ -52,12 +52,20 @@ typedef enum {
 	GOG_AXIS_TICK_MINOR
 } GogAxisTickTypes;
 
+typedef struct {
+	double		 position;
+	GogAxisTickTypes type;
+	GOString        *str;
+} GogAxisTick;
+
 #define GOG_TYPE_AXIS_LINE	(gog_axis_line_get_type ())
 #define GOG_AXIS_LINE(o)	(G_TYPE_CHECK_INSTANCE_CAST ((o), GOG_TYPE_AXIS_LINE, GogAxisLine))
 #define GOG_IS_AXIS_LINE(o)	(G_TYPE_CHECK_INSTANCE_TYPE ((o), GOG_TYPE_AXIS_LINE))
 
 GType gog_axis_line_get_type (void);
 
+unsigned	 gog_axis_line_get_ticks (GogAxisLine *axis_line, GogAxisTick **ticks);
+
 G_END_DECLS
 
 #endif /*GOG_AXIS_BASE_H*/
